/********************************************************/
/* Copyright (C) 2001 OSS Nokalva.  All rights reserved.*/
/********************************************************/

/* THIS FILE IS PROPRIETARY MATERIAL OF OSS Nokalva, INC. AND
 * MAY BE USED ONLY BY DIRECT LICENSEES OF OSS Nokalva, INC.
 * THIS FILE MAY NOT BE DISTRIBUTED. */

/* Prepared by OSS Nokalva, Inc.  07/22/2001      */

/**
    Application program Thuge.java
    Demonstrates how one can avoid intermediate disk I/O when encoding huge
    container. Consider the following abstract syntax:

    MeasDataCollection ::= SEQUENCE {
	measFileHeader MeasFileHeader,
	measData SEQUENCE --<ValueInFile>-- OF BBCard,
	measFileFooter MeasFileFooter
    }

    where MeasDataCollection.measData can contain a huge number of elements
    and the number of elements to encode is not known in advance. Normally, an
    application will take the following steps to encode the MeasDataCollection:

    // 1. Create and fill the MeasDataCollection.measData container. The code
    //    assumes that the elements of the container are not already in the
    //    serialized form but are generated by the application.
    OSSObjectStorage objects =
	OSSStorageManager.getInstance().allocate(StorageManager.STORAGE_OBJECTS);
    ObjectWriter writer = null;
    try {
	writer = objects.getWriter(false);
	while (<has more elements>)
	    writer.writeObject(...);
    } finally {
	writer.close();
    }

    // 2. Construct the instance of MeasDataCollection
    MeasDataCollection data = new MeasDataCollection(
	new MeasFileHeader(...),
	new MeasDataCollection.MeasData(objects),
	new MeasFileFooter(...)
    );

    // 3. Pass the value to the encoder.
    Coder coder = Meas.getBERCoder();
    FileOutputStream sink = new FileOutputStream("meas.ber");
    coder.encode(data, sink);
    sink.close();

    The obvious drawback of this approach is that it saves the elements,
    generated by the application, into the intermediate disk file before
    encoding and that slows down the application performance. For brevity, we will
    refer to the code that generates elements as the "Provider". Note, that
    in the above example the "Provider" works as "master", i.e. it explicitly
    fills the ObjectStorage with elements.

    This sample program demonstrates a better approach that avoids intermediate disk
    I/O by reversing the roles of the ObjectStorage and "Provider". The
    ObjectStorage becomes the "master" and the "Provider" works as the "slave".
    In other words, ObjectStorage remains empty until the coder will emit a
    read request. Then the ObjectStorage iteratively activates the "Provider"
    to get the elements to encode directly from the "Provider" thus avoiding
    an intermediate disk file.

    The sample defines three classes:
    - The Thuge class is the main application class. It initializes the runtime,
      instantiates the coder, creates the value of MeasDataCollection and invokes
      the encoder to encode the value into the disk file.
    - The DataProvider class simulates the application code by generating
      elements for the huge container.
    - The LazyStorage class is the custom implementation of the ObjectStorage
      interface, that implements lazy "loading" of the storage. Or, in other
      words, that implements the pipeline between the "Provider" and the
      encoder to avoid any intermediate disk files.

    Although the approach with LazyStorage works faster, it has one obvious
    drawback: if some encoder's error will occur in the middle of that huge
    SEQUENCE OF, the elements, processed thus far, will be lost. And the
    application will have to restart the encoding from the very beginning,
    requesting the "Provider" to resend all the elements once again. To address
    this issue, the sample also demonstrates how you can monitor the process
    of encoding and do limited error recovery.

    When encoding a huge SEQUENCE OF, it's most likely that the encoder's error can
    occur when encoding an individual element of the container. In order to
    intercept a possible encoder error, this sample modifies the original
    ASN.1 definition:

    MeasDataCollection ::= SEQUENCE {
	measFileHeader MeasFileHeader,
	measData SEQUENCE --<ValueInFile>-- OF TYPE-IDENTIFIER.&Type(BBCard),
	measFileFooter MeasFileFooter
    }

    Note, that the type of the element is changed from the BBCard to the open type.
    The type constraint, imposed on open type, says that the open type always
    carries the value of BBCard type. As long as you use BER or DER, this new
    definition is equivalent to the original one.

    Having an open type as the element of the container, you can perform the
    encoding of an individual element in your code thus having complete control
    on the exception handling. Once you manually encoded an open type, the
    runtime will simply concatenate the encodings of individual elements to
    build the resulting encoding.
*/

/* Compiler-generated classes */
import meas.*;
import meas.bcas.*;

/* Universal classes from the OSS runtime library */
import com.oss.asn1.*;

/* Java I/O and utility classes */
import java.io.*;
import java.util.NoSuchElementException;

public class Thuge {
    public static void main(String[] args)
    {
	// Initialize the project
	try {
	    Meas.initialize();
	} catch (Exception e) {
	    System.out.println("MEAS initialization error: " + e);
	    System.exit(1);
	}

	// Instantiate and configure BER coder
	Coder coder = Meas.getBERCoder();
	coder.enableEncoderDebugging();
	coder.enableDecoderDebugging();
	coder.useIndefiniteLengthEncoding();

	// Instantiate the object that will provide elements for the
	// MeasDataCollection.measData container
	DataProvider dataProvider = new DataProvider(100);

	// Create the value of MeasDataCollection to encode
	MeasDataCollection collection = new MeasDataCollection(
		new MeasFileHeader("Data file header"),
		new MeasDataCollection.MeasData(new LazyStorage(dataProvider)),
		new MeasFileFooter("Data file footer")
	);

	// Encode the value and save the encoding into the "meas.data" file.
	try {
	    FileOutputStream blob = new FileOutputStream("meas.data");
	    coder.encode(collection, blob);
	    blob.close();
	} catch (Exception e) {
	    System.out.println("Encoder failed: " + e);
	}
	Meas.deinitialize();
    }
}

// Sample class that implements a "Provider" operating in the slave mode.
class DataProvider {

    int mCount = 0;
    int mCurrentElement = 0;

    public DataProvider(int count)
    {
	mCount = count;
    }

    public int getSize()
    {
	return mCount;
    }

    public Object nextElement()
    {
	if (mCurrentElement < mCount) {
	    BBCard element = (BBCard)BCAS.myCard.clone();
	    if (mCurrentElement % 10 == 0)
		// Generate some illegal elements to demonstrate how the
		// error recovery does work
		element.setAge(101);
	    else
		element.setAge(mCurrentElement % 100 + 1);
	    ++mCurrentElement;
	    return element;
	} else
	    throw new NoSuchElementException("element" + mCurrentElement);
    }

    public boolean hasMoreElements()
    {
	return (mCurrentElement < mCount);
    }
}

// Custom implementation of the ObjectStorage interface that requests elements
// to be encoded from the "Provider".
class LazyStorage implements ObjectStorage {

    // The "Provider" that will supply elements.
    DataProvider mProvider = null;

    // Construct the LazyStorage and attach it to the "Provider" that will
    // supply the elements.
    public LazyStorage(DataProvider provider)
    {
	mProvider = provider;
    }

    // The methods below implement the ObjectStorage interface.
    public int getKind()
    {
	return StorageManager.STORAGE_OBJECTS;
    }

    // Indicate that this kind of storage is readable.
    public boolean canRead()
    {
	return true;
    }

    // Indicate that this kind of storage is read-only.
    public boolean canWrite()
    {
	return false;
    }

    // Return the number of elements in the storage. This method returns -1
    // to indicate that the number of elements in the container is not known
    // in advance (the indefinite-length case).
    public long getSize()
    {
	return -1;
    }

    // LazyStorage does not consume any resources that need explicit
    // destruction. That's why the deallocate() is a dummy method.
    public void deallocate()
    {
    }

    // Dummy method for now.
    public Storage copy()
    {
	throw new StorageException("copy is not implemented");
    }

    // The reset() and getWriter() should throw a StorageException
    // when the storage is read-only.
    public void reset(boolean zeroize)
    {
	throw new StorageException("the Storage is read-only");
    }

    public ObjectWriter getWriter(boolean append)
    {
	throw new StorageException("the Storage is read-only");
    }

    // Return the ObjectReader to provide elements to the caller. The
    // ObjectReader is implemented as anonymous inner class that is a simple
    // wrapper class to the "Provider". The class works as follows: its
    // hasMoreObject() method iteratively requests elements from the
    // "Provider" until it gets a valid element that can be successively
    // encoded. Then it converts the element to the OpenType and saves it
    // for further retrieval by the readObject().
    public ObjectReader getReader()
    {
	return new ObjectReader() {
	    // This coder is used to encode individual elements.
	    Coder mCoder = Meas.getBERCoder();
	    // The last "good" element delivered by the "Provider"
	    OpenType mNext = null;

	    // Request elements from the provider. Skip illegal element(s),
	    // that fail to encode.
	    public boolean hasMoreObjects()
	    {
		while (mNext == null) {
		    if (!mProvider.hasMoreElements())
			break;
		    // The provider still has some elements to encode.
		    AbstractData element =
			(AbstractData)mProvider.nextElement();
		    // Try to encode the next element.
		    try {
			// Save the encoding for future retrieval by the
			// call to readObject()
			mNext = new OpenType(element, mCoder);
		    } catch (Exception e) {
			// The "Provider" returned an illegal element.
			// Do appropriate error recovery here. This
			// sample simply prints the bad element to stderr.
			System.err.println("Problem with element: " +
			    element + " - " + e);
		    }
		}
		return mNext != null;
	    }
	    // Return next object to encode.
	    public Object readObject()
	    {
		try {
		    if (hasMoreObjects()) {
			return mNext;
		    } else
			return null;
		} finally {
		    mNext = null;
		}
	    }
	    public void close()
	    {
	    }
	};
    }
}
